from typing import Optional

from pydbsp.lazy_zset import LazyZSet
from pydbsp.lazy_zset.functions.linear import Cmp, Projection, project, select
from pydbsp.stream import Lift1, Stream, StreamHandle, step_until_fixpoint_and_return


class LiftedSelect[T](Lift1[LazyZSet[T], LazyZSet[T]]):
    def __init__(self, stream: Optional[StreamHandle[LazyZSet[T]]], p: Cmp[T]):
        super().__init__(stream, lambda z: select(z, p), None)


class LiftedLiftedSelect[T](Lift1[Stream[LazyZSet[T]], Stream[LazyZSet[T]]]):
    def __init__(self, stream: Optional[StreamHandle[Stream[LazyZSet[T]]]], p: Cmp[T]):
        super().__init__(
            stream,
            lambda x: step_until_fixpoint_and_return(LiftedSelect(StreamHandle(lambda: x), p)),
            None,
        )


class LiftedProject[T, R](Lift1[LazyZSet[T], LazyZSet[R]]):
    def __init__(self, stream: Optional[StreamHandle[LazyZSet[T]]], f: Projection[T, R]):
        super().__init__(stream, lambda z: project(z, f), None)


class LiftedLiftedProject[T, R](Lift1[Stream[LazyZSet[T]], Stream[LazyZSet[R]]]):
    def __init__(self, stream: Optional[StreamHandle[Stream[LazyZSet[T]]]], f: Projection[T, R]):
        super().__init__(
            stream,
            lambda x: step_until_fixpoint_and_return(LiftedProject(StreamHandle(lambda: x), f)),
            None,
        )



from typing import Optional, TypeVar

from pydbsp.core import AbelianGroupOperation
from pydbsp.stream import Stream, StreamHandle, UnaryOperator
from pydbsp.stream.operators.linear import Differentiate, Integrate, LiftedDelay, LiftedIntegrate
from pydbsp.zset import ZSet
from pydbsp.zset.operators.binary import LiftedLiftedH

T = TypeVar("T")


class DeltaLiftedDeltaLiftedDistinct(UnaryOperator[Stream[ZSet[T]], Stream[ZSet[T]]]):
    """
    An operator for incrementally maintaining distinct elements in a stream of
    streams. See :func:`~pydbsp.zset.functions.binary.H`
    """

    integrated_diff_stream_a: Integrate[Stream[ZSet[T]]]
    lift_integrated_diff_stream_a: LiftedIntegrate[ZSet[T]]
    lift_delay_lift_integrated_diff_stream_a: LiftedDelay[ZSet[T]]
    lift_lift_H: LiftedLiftedH[T]
    diff_lift_lift_H: Differentiate[Stream[ZSet[T]]]

    def set_input(
        self,
        stream_handle: StreamHandle[Stream[ZSet[T]]],
        output_stream_group: Optional[AbelianGroupOperation[Stream[ZSet[T]]]],
    ) -> None:
        self.input_stream_handle = stream_handle
        self.integrated_diff_stream_a = Integrate(self.input_stream_handle)
        self.lift_integrated_diff_stream_a = LiftedIntegrate(self.integrated_diff_stream_a.output_handle())
        self.lift_delay_lift_integrated_diff_stream_a = LiftedDelay(self.lift_integrated_diff_stream_a.output_handle())
        self.lift_lift_H = LiftedLiftedH(
            self.integrated_diff_stream_a.output_handle(),
            self.lift_delay_lift_integrated_diff_stream_a.output_handle(),
        )
        self.diff_lift_lift_H = Differentiate(self.lift_lift_H.output_handle())  # type: ignore
        self.output_stream_handle = self.diff_lift_lift_H.output_handle()

    def __init__(self, diff_stream_a: Optional[StreamHandle[Stream[ZSet[T]]]]):
        super().__init__(diff_stream_a, None)

    def step(self) -> bool:
        fixedpoint_integrated_diff_stream_a = self.integrated_diff_stream_a.step()
        fixedpoint_lift_integrated_diff_stream_a = self.lift_integrated_diff_stream_a.step()
        fixedpoint_lift_delay_lift_integrated_diff_stream_a = self.lift_delay_lift_integrated_diff_stream_a.step()
        fixedpoint_lift_lift_H = self.lift_lift_H.step()
        fixedpoint_diff_lift_lift_H = self.diff_lift_lift_H.step()

        return (
            fixedpoint_integrated_diff_stream_a
            and fixedpoint_lift_integrated_diff_stream_a
            and fixedpoint_lift_delay_lift_integrated_diff_stream_a
            and fixedpoint_lift_lift_H
            and fixedpoint_diff_lift_lift_H
        )



from typing import Callable, Dict, TypeVar

from pydbsp.zset import ZSet

T = TypeVar("T")

Cmp = Callable[[T], bool]


def select[T](zset: ZSet[T], p: Cmp[T]) -> ZSet[T]:
    """Filters a Z-set based on a predicate function."""
    return ZSet({k: v for k, v in zset.items() if p(k)})


R = TypeVar("R")
Projection = Callable[[T], R]


def project[T, R](zset: ZSet[T], f: Projection[T, R]) -> ZSet[R]:
    """
    Projects a Z-set to a new Z-set by applying a function to each element.

    Returns a new Z-set where each element is the result of applying f to an element
    of the input Z-set. Weights of elements mapping to the same value are summed.
    """
    output: Dict[R, int] = {}
    for value, weight in zset.items():
        fvalue = f(value)
        if fvalue not in output:
            output[fvalue] = weight
        else:
            output[fvalue] += weight

    return ZSet(output)



from pydbsp.lazy_zset import LazyZSet
from pydbsp.lazy_zset.functions.binary import H
from pydbsp.stream import Lift2, Stream, StreamHandle, step_until_fixpoint_and_return


class LiftedH[T](Lift2[LazyZSet[T], LazyZSet[T], LazyZSet[T]]):
    def __init__(
        self,
        diff_stream_a: StreamHandle[LazyZSet[T]],
        integrated_stream_a: StreamHandle[LazyZSet[T]],
    ):
        super().__init__(diff_stream_a, integrated_stream_a, H, None)


class LiftedLiftedH[T](Lift2[Stream[LazyZSet[T]], Stream[LazyZSet[T]], Stream[LazyZSet[T]]]):
    def __init__(
        self,
        integrated_diff_stream_a: StreamHandle[Stream[LazyZSet[T]]],
        lifted_delayed_lifted_integrated_stream_a: StreamHandle[Stream[LazyZSet[T]]],
    ):
        super().__init__(
            integrated_diff_stream_a,
            lifted_delayed_lifted_integrated_stream_a,
            lambda x, y: step_until_fixpoint_and_return(LiftedH(StreamHandle(lambda: x), StreamHandle(lambda: y))),
            None,
        )



from itertools import product
from typing import Callable, TypeVar

from pydbsp.indexed_zset import IndexedZSet, sort_merge_join
from pydbsp.zset import ZSet

I = TypeVar("I")
T = TypeVar("T")
R = TypeVar("R")
S = TypeVar("S")
PostSortMergeJoinProjection = Callable[[I, T, R], S]


def join_with_index[T, I, R, S](
    left_indexed_zset: IndexedZSet[I, T],
    right_indexed_zset: IndexedZSet[I, R],
    f: PostSortMergeJoinProjection[I, T, R, S],
) -> ZSet[S]:
    """
    Joins two ZSets. Takes advantage of the B-Tree indexes and is implemented as a sort-merge join.

    Args:
       left_zset
       right_zset
       p: Join key function
       f: projection to be applied to the join
    """

    output: ZSet[S] = ZSet({})

    for match in sort_merge_join(left_indexed_zset.index, right_indexed_zset.index):
        left_values = [(value, left_indexed_zset.inner[value]) for value in left_indexed_zset.index_to_value[match]]
        right_values = [(value, right_indexed_zset.inner[value]) for value in right_indexed_zset.index_to_value[match]]
        left_x_right = product(left_values, right_values)

        projected_values = [
            (f(match, left_value, right_value), left_weight * right_weight)
            for ((left_value, left_weight), (right_value, right_weight)) in left_x_right
        ]

        for projected_value, new_weight in projected_values:
            if projected_value in output:
                output[projected_value] += new_weight
            else:
                output[projected_value] = new_weight

    return output



from typing import Callable, Dict, TypeVar

from pydbsp.zset import ZSet

T = TypeVar("T")
R = TypeVar("R")
S = TypeVar("S")
JoinCmp = Callable[[T, R], bool]
PostJoinProjection = Callable[[T, R], S]


def join[T, R, S](
    left_zset: ZSet[T],
    right_zset: ZSet[R],
    p: JoinCmp[T, R],
    f: PostJoinProjection[T, R, S],
) -> ZSet[S]:
    """
    Joins two ZSets. Implemented as a nested loop join.

    Args:
       left_zset
       right_zset
       p: Join key function
       f: projection to be applied to the join
    """
    output: Dict[S, int] = {}
    for left_value, left_weight in left_zset.items():
        for right_value, right_weight in right_zset.items():
            if p(left_value, right_value):
                projected_value = f(left_value, right_value)
                new_weight = left_weight * right_weight

                if projected_value in output:
                    output[projected_value] += new_weight
                else:
                    output[projected_value] = new_weight

    return ZSet(output)



from abc import abstractmethod
from typing import Protocol, TypeVar

T = TypeVar("T")


class AbelianGroupOperation(Protocol[T]):
    """
    This protocol defines the basic operations and properties of an Abelian group,
    addition and negation.
    """

    @abstractmethod
    def add(self, a: T, b: T) -> T:
        raise NotImplementedError

    @abstractmethod
    def neg(self, a: T) -> T:
        """Returns the inverse of the given element"""
        raise NotImplementedError

    @abstractmethod
    def identity(self) -> T:
        """Returns the identity, zero, element of the group"""
        raise NotImplementedError

    def is_commutative(self, a: T, b: T) -> bool:
        """
        Returns whether a + b == b + a.
        """
        test = self.add(a, b) == self.add(b, a)
        if not test:
            print(f"Failed commutativity assertion: {self.add(a, b)} == {self.add(b, a)}")

        return test

    def is_associative(self, a: T, b: T, c: T) -> bool:
        """
        Returns whether (a + b) + c == a + (b + c).
        """
        test = self.add(self.add(a, b), c) == self.add(a, self.add(b, c))
        if not test:
            print(f"Failed associativity assertion: {self.add(self.add(a, b), c)} == {self.add(a, self.add(b, c))}")

        return test

    def has_identity(self, a: T) -> bool:
        """
        Checks if the identity element behaves correctly for the given element.
        """
        identity = self.identity()
        test = self.add(a, identity) == a and self.add(identity, a) == a
        if not test:
            print(f"Failed identity assertion: {self.add(a, identity)} == {self.add(identity, a)}")

        return test

    def has_inverse(self, a: T) -> bool:
        """
        Returns if the given element has a well defined inverse.
        """
        identity = self.identity()
        inv_a = self.neg(a)
        test = self.add(a, inv_a) == identity and self.add(inv_a, a) == identity
        if not test:
            print(f"Failed inverse assertion: {self.add(a, inv_a)} == {self.add(inv_a, a)}")

        return test



from typing import Dict

from pydbsp.zset import ZSet


def H[T](diff: ZSet[T], integrated_state: ZSet[T]) -> ZSet[T]:
    """
    Computes a "distincted" difference between a diff and an integrated state.

    Args:
        diff (ZSet[T]): The difference ZSet, representing changes.
        integrated_state (ZSet[T]): The current integrated state ZSet.

    Returns:
        ZSet[T]: A new ZSet representing the "distincted" difference.

    Behavior:
    1. For each element k in the diff:
       a. If k is not in the integrated_state and has a positive weight in diff,
          it's added to the result with weight 1.
       b. If k is in the integrated_state:
          - If its weight becomes non-positive after applying the diff and was
            previously positive, it's added to the result with weight -1.
          - If its weight becomes positive after applying the diff and was
            previously non-positive, it's added to the result with weight 1.
    2. All other cases result in the element not being included in the output.

    This function effectively tracks which elements are being added to or removed from
    a distinct set representation, based on the integrated state and the incoming changes.
    """
    distincted_diff: Dict[T, int] = {}
    for k, v in diff.items():
        current_k_latest_diff_weight = v

        if k in integrated_state:
            current_k_latest_delayed_state_weight = integrated_state[k]
            coalesced_weight = current_k_latest_diff_weight + current_k_latest_delayed_state_weight

            if current_k_latest_delayed_state_weight > 0 and coalesced_weight <= 0:
                distincted_diff[k] = -1

                continue

            if current_k_latest_delayed_state_weight <= 0 and coalesced_weight > 0:
                distincted_diff[k] = 1
        else:
            if v > 0:
                distincted_diff[k] = 1

    return ZSet(distincted_diff)



from pydbsp.indexed_zset import IndexedZSet, Indexer
from pydbsp.zset import ZSet


def index_zset[I, T](zset: ZSet[T], indexer: Indexer[T, I]) -> IndexedZSet[I, T]:
    """Indexes a Z-Set according to some indexer function."""
    return IndexedZSet(zset.inner, indexer)



from typing import Any, Dict, List, NewType, Optional, Set, Tuple, TypeAlias, TypeVar, cast

from pydbsp.indexed_zset.operators.linear import LiftedLiftedIndex
from pydbsp.stream import (
    BinaryOperator,
    Lift1,
    LiftedGroupAdd,
    Stream,
    StreamHandle,
    step_until_fixpoint_and_return,
)
from pydbsp.stream.operators.linear import (
    Delay,
    LiftedStreamElimination,
    LiftedStreamIntroduction,
)
from pydbsp.zset import ZSet, ZSetAddition
from pydbsp.zset.operators.bilinear import DeltaLiftedDeltaLiftedJoin
from pydbsp.zset.operators.unary import DeltaLiftedDeltaLiftedDistinct

Constant = Any
_Variable: TypeAlias = str
Variable = NewType("Variable", _Variable)
Term = Constant | Variable

Predicate = str
Atom = Tuple[Predicate, Tuple[Term, ...]]
Fact = Tuple[Predicate, Tuple[Constant, ...]]
Rule = Tuple[Atom, ...]

Program = ZSet[Rule]
EDB = ZSet[Fact]


class Rewrite:
    inner: Dict[Variable, Constant]

    def __init__(self, substitutions: Dict[Variable, Constant]) -> None:
        self.inner = substitutions

    def __repr__(self) -> str:
        return self.inner.__repr__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Rewrite):
            return False

        return self.inner == other.inner  # type: ignore

    def __contains__(self, variable: Variable) -> bool:
        return self.inner.__contains__(variable)

    def __getitem__(self, variable: Variable) -> Constant | None:
        if variable not in self.inner:
            return None

        return self.inner[variable]

    def __hash__(self) -> int:
        items = sorted(list(self.inner.items()), key=lambda x: x[0])

        return hash(tuple(items))

    def apply(self, atom: Atom) -> Atom:
        terms: List[Term] = list(atom[1])

        for idx, term in enumerate(atom[1]):
            if term in self:
                terms[idx] = self[term]

        new_atom: Atom = (atom[0], tuple(terms))

        return new_atom


class RewriteMonoid:
    def add(self, a: Rewrite, b: Rewrite) -> Rewrite:
        merge = self.identity()
        for var, constant in a.inner.items():
            key = Variable(var)

            if key not in merge:
                merge.inner[key] = constant

        for var, constant in b.inner.items():
            key = Variable(var)

            if key not in merge:
                merge.inner[key] = constant

        return merge

    def identity(self) -> Rewrite:
        return Rewrite({})

    def is_associative(self, a: Rewrite, b: Rewrite, c: Rewrite) -> bool:
        return self.add(self.add(a, b), c) == self.add(a, self.add(b, c))


rewrite_monoid = RewriteMonoid()

RewriteSet = ZSet[Rewrite]


def unify(atom: Atom, fact: Fact) -> Rewrite | None:
    unification_result = rewrite_monoid.identity()

    atom_term_count = len(atom[1])
    fact_constant_count = len(fact[1])
    if atom_term_count != fact_constant_count:
        return rewrite_monoid.identity()

    for left_term, right_constant in zip(atom[1], fact[1]):
        if isinstance(left_term, _Variable):
            if left_term not in unification_result:
                substitution: Dict[Variable, Constant] = {left_term: right_constant}

                unification_result = rewrite_monoid.add(unification_result, Rewrite(substitution))
            else:
                left_variable_bound_constant = unification_result[left_term]
                cast(Constant, left_variable_bound_constant)

                if left_variable_bound_constant != right_constant:
                    return None
        else:
            if left_term != right_constant:
                return None

    return unification_result


T = TypeVar("T")
R = TypeVar("R")
S = TypeVar("S")
Q = TypeVar("Q")

Provenance: TypeAlias = int
Direction: TypeAlias = tuple[Provenance, Provenance, Atom | None]
ProvenanceChain: TypeAlias = ZSet[Direction]

Head: TypeAlias = Atom
Signal = tuple[Provenance, Head]
GroundingSignals = ZSet[Signal]

EmptyGroundingCanaries: GroundingSignals = ZSetAddition().identity()


def sig(program: Program) -> GroundingSignals:
    signals: GroundingSignals = ZSetAddition().identity()

    for rule, weight in program.items():
        running_provenance = 0

        if len(rule) > 1:
            for dependency in rule[1:]:
                running_provenance += hash(dependency)

        full_provenance = running_provenance
        signals.inner[(full_provenance, rule[0])] = weight  # type: ignore

    return signals


class LiftedSig(Lift1[Program, GroundingSignals]):
    def __init__(self, stream: Optional[StreamHandle[Program]]):
        group: ZSetAddition[Signal] = ZSetAddition()

        super().__init__(stream, sig, group)


def dir(program: Program) -> ProvenanceChain:
    directions: ProvenanceChain = ZSetAddition().identity()

    for rule, weight in program.items():
        running_provenance = 0

        if len(rule) > 1:
            for dependency in rule[1:]:
                previous_atom_provenance = running_provenance
                running_provenance += hash(dependency)

                directions.inner[(previous_atom_provenance, running_provenance, dependency)] = weight

    return directions


class LiftedDir(Lift1[Program, ProvenanceChain]):
    def __init__(self, stream: Optional[StreamHandle[Program]]):
        group: ZSetAddition[Direction] = ZSetAddition()

        super().__init__(stream, dir, group)


AtomSet: TypeAlias = ZSet[Atom]
ProvenanceIndexedRewrite = Tuple[Provenance, Rewrite]
AtomWithSourceRewriteAndProvenance = Tuple[Provenance, Atom | None, Rewrite]


def rewrite_product_projection(
    atom_rewrite_provenance: AtomWithSourceRewriteAndProvenance, fact: Fact
) -> Tuple[Provenance, Rewrite]:
    provenance = atom_rewrite_provenance[0]
    possible_atom = atom_rewrite_provenance[1]
    rewrite = atom_rewrite_provenance[2]

    monoid = RewriteMonoid()
    fresh_rewrite = monoid.identity()

    if possible_atom is not None:
        fresh_rewrite = cast(Rewrite, unify(rewrite.apply(possible_atom), fact))

    new_rewrite = monoid.add(rewrite, fresh_rewrite)

    return provenance, new_rewrite


edb_identity: EDB = ZSetAddition().identity()
provenance_indexed_rewrite_identity: ProvenanceChain = ZSetAddition().identity()


class IncrementalDatalog(BinaryOperator[EDB, Program, EDB]):
    # Program transformations
    lifted_sig: LiftedSig
    lifted_intro_lifted_sig: LiftedStreamIntroduction[GroundingSignals]
    lifted_dir: LiftedDir
    lifted_intro_lifted_dir: LiftedStreamIntroduction[ProvenanceChain]

    # EDB transformations
    lifted_intro_edb: LiftedStreamIntroduction[EDB]

    # Rewrite transformations
    rewrites: StreamHandle[ZSet[ProvenanceIndexedRewrite]]
    lifted_rewrites: LiftedStreamIntroduction[ZSet[ProvenanceIndexedRewrite]]

    # Joins
    gatekeep: DeltaLiftedDeltaLiftedJoin[ProvenanceIndexedRewrite, Direction, AtomWithSourceRewriteAndProvenance]
    product: DeltaLiftedDeltaLiftedJoin[AtomWithSourceRewriteAndProvenance, Fact, ProvenanceIndexedRewrite]
    ground: DeltaLiftedDeltaLiftedJoin[ProvenanceIndexedRewrite, Signal, Fact]

    # Distincts
    distinct_rewrites: DeltaLiftedDeltaLiftedDistinct[ProvenanceIndexedRewrite]
    distinct_facts: DeltaLiftedDeltaLiftedDistinct[Fact]

    # Delays
    delay_distinct_facts: Delay[Stream[ZSet[Fact]]]
    delay_distinct_rewrites: Delay[Stream[ZSet[ProvenanceIndexedRewrite]]]

    # Pluses
    fresh_facts_plus_edb: LiftedGroupAdd[Stream[EDB]]
    rewrite_product_plus_rewrites: LiftedGroupAdd[Stream[ZSet[ProvenanceIndexedRewrite]]]

    # Stream elimination
    lifted_elim_fresh_facts: LiftedStreamElimination[EDB]

    def set_input_a(self, stream_handle_a: StreamHandle[EDB]) -> None:
        self.input_stream_handle_a = stream_handle_a
        self.lifted_intro_edb = LiftedStreamIntroduction(self.input_stream_handle_a)

        provenance_indexed_rewrite_group: ZSetAddition[ProvenanceIndexedRewrite] = ZSetAddition()
        rewrite_stream: Stream[ZSet[ProvenanceIndexedRewrite]] = Stream(provenance_indexed_rewrite_group)
        self.rewrites = StreamHandle(lambda: rewrite_stream)
        empty_rewrite_set = rewrite_stream.group().identity()
        empty_rewrite_set.inner[(0, RewriteMonoid().identity())] = 1

        self.rewrites.get().send(empty_rewrite_set)
        self.lifted_rewrites = LiftedStreamIntroduction(self.rewrites)

    def set_input_b(self, stream_handle_b: StreamHandle[Program]) -> None:
        self.input_stream_handle_b = stream_handle_b

        self.lifted_sig = LiftedSig(self.input_stream_handle_b)
        self.lifted_intro_lifted_sig = LiftedStreamIntroduction(self.lifted_sig.output_handle())

        self.lifted_dir = LiftedDir(self.input_stream_handle_b)
        self.lifted_intro_lifted_dir = LiftedStreamIntroduction(self.lifted_dir.output_handle())

        self.gatekeep = DeltaLiftedDeltaLiftedJoin(
            None, None, lambda left, right: left[0] == right[0], lambda left, right: (right[1], right[2], left[1])
        )

        self.product = DeltaLiftedDeltaLiftedJoin(
            self.gatekeep.output_handle(),
            None,
            lambda left, right: left[1] is None
            or (left[1][0] == right[0] and unify(left[2].apply(left[1]), right) is not None),
            rewrite_product_projection,
        )

        self.ground = DeltaLiftedDeltaLiftedJoin(
            self.product.output_handle(),
            self.lifted_intro_lifted_sig.output_handle(),
            lambda left, right: left[0] == right[0],
            lambda left, right: left[1].apply(right[1]),
        )

        self.fresh_facts_plus_edb = LiftedGroupAdd(self.ground.output_handle(), self.lifted_intro_edb.output_handle())
        self.distinct_facts = DeltaLiftedDeltaLiftedDistinct(self.fresh_facts_plus_edb.output_handle())

        self.rewrite_product_plus_rewrites = LiftedGroupAdd(
            self.product.output_handle(), self.lifted_rewrites.output_handle()
        )
        self.distinct_rewrites = DeltaLiftedDeltaLiftedDistinct(self.rewrite_product_plus_rewrites.output_handle())

        self.delay_distinct_facts = Delay(self.distinct_facts.output_handle())
        self.delay_distinct_rewrites = Delay(self.distinct_rewrites.output_handle())
        self.gatekeep.set_input_a(self.delay_distinct_rewrites.output_handle())
        self.gatekeep.set_input_b(self.lifted_intro_lifted_dir.output_handle())
        self.product.set_input_b(self.delay_distinct_facts.output_handle())

        self.lifted_elim_fresh_facts = LiftedStreamElimination(self.distinct_facts.output_handle())
        self.output_stream_handle = self.lifted_elim_fresh_facts.output_handle()

    def step(self) -> bool:
        self.lifted_intro_edb.step()
        self.lifted_rewrites.step()
        self.lifted_sig.step()
        self.lifted_dir.step()
        self.lifted_intro_lifted_sig.step()
        self.lifted_intro_lifted_dir.step()
        self.gatekeep.step()
        self.product.step()
        self.ground.step()
        self.fresh_facts_plus_edb.step()
        self.distinct_facts.step()
        self.rewrite_product_plus_rewrites.step()
        self.distinct_rewrites.step()
        self.delay_distinct_facts.step()
        self.delay_distinct_rewrites.step()
        self.lifted_elim_fresh_facts.step()

        new_facts_stream = self.distinct_facts.output().latest()
        new_rewrites_stream = self.distinct_rewrites.output().latest()
        return new_facts_stream.is_identity() and new_rewrites_stream.is_identity()


ColumnReference = Tuple[int, ...]
JoinIndex = Tuple[Predicate, ColumnReference]
NaiveJoinOrder = ZSet[JoinIndex]


def compute_rule_index(rule: Rule) -> NaiveJoinOrder:
    group: ZSetAddition[JoinIndex] = ZSetAddition()
    column_index = group.identity()

    variables: Set[Variable] = set()
    fresh_variables: Set[Variable] = set()

    for body_atom in rule[1:]:
        predicate = body_atom[0]
        columns: List[int] = []

        for idx, term in enumerate(body_atom[1]):
            if isinstance(term, _Variable) and term not in variables:
                fresh_variables.add(term)

                continue

            columns.append(idx)

        column_index.inner[(predicate, tuple(columns))] = 1
        for variable in fresh_variables:
            variables.add(variable)

        fresh_variables.clear()

    return column_index


def jorder(program: Program) -> NaiveJoinOrder:
    group: ZSetAddition[JoinIndex] = ZSetAddition()
    column_index = group.identity()

    for rule, _weight in program.items():
        rule_index = compute_rule_index(rule)

        for index_schema, _weight in rule_index.items():
            column_index.inner[index_schema] = 1

    return column_index


class LiftedJorder(Lift1[Program, NaiveJoinOrder]):
    def __init__(self, stream: Optional[StreamHandle[Program]]):
        super().__init__(stream, lambda p: jorder(p), None)


class LiftedLiftedJorder(Lift1[Stream[Program], Stream[NaiveJoinOrder]]):
    def __init__(
        self,
        stream: Optional[StreamHandle[Stream[Program]]],
    ):
        super().__init__(
            stream,
            lambda sp: step_until_fixpoint_and_return(LiftedJorder(StreamHandle(lambda: sp))),
            None,
        )


IndexedFact = Tuple[Fact, Fact]


def index_fact(column_reference: ColumnReference, fact: Fact) -> IndexedFact:
    indexed_fact_terms: List[int] = []
    for column in column_reference:
        indexed_fact_terms.append(fact[1][column])

    frozen_indexed_fact_terms = tuple(indexed_fact_terms)
    if len(frozen_indexed_fact_terms) == 0:
        frozen_indexed_fact_terms = fact[1]

    return ((fact[0], frozen_indexed_fact_terms), fact)


ConstantTerms = Tuple[Constant, ...]


def get_constant_terms(atom: Atom) -> ConstantTerms:
    constant_terms: List[Constant] = []
    for term in atom[1]:
        if not isinstance(term, _Variable):
            constant_terms.append(term)

    return tuple(constant_terms)


ExtendedProvenanceDirection: TypeAlias = tuple[Provenance, Provenance, Atom | None, ColumnReference]
ExtendedProvenanceChain: TypeAlias = ZSet[ExtendedProvenanceDirection]

ColumnReferenceSequence = List[ColumnReference]


AtomWithSourceRewriteAndExtendedProvenance = Tuple[Provenance, Atom | None, Rewrite, ColumnReference]


class IncrementalDatalogWithIndexing(BinaryOperator[EDB, Program, EDB]):
    # Program transformations
    lifted_sig: LiftedSig
    lifted_intro_lifted_sig: LiftedStreamIntroduction[GroundingSignals]
    lifted_dir: LiftedDir
    lifted_intro_lifted_dir: LiftedStreamIntroduction[ProvenanceChain]
    lifted_jorder: LiftedJorder
    lifted_intro_lifted_jorder: LiftedStreamIntroduction[NaiveJoinOrder]

    # EDB transformations
    lift_intro_edb: LiftedStreamIntroduction[EDB]

    # Rewrite transformations
    rewrites: StreamHandle[ZSet[ProvenanceIndexedRewrite]]
    lift_rewrites: LiftedStreamIntroduction[ZSet[ProvenanceIndexedRewrite]]

    # Joins
    gatekeep: DeltaLiftedDeltaLiftedJoin[ProvenanceIndexedRewrite, Direction, AtomWithSourceRewriteAndProvenance]
    product: DeltaLiftedDeltaLiftedJoin[AtomWithSourceRewriteAndProvenance, IndexedFact, ProvenanceIndexedRewrite]
    ground: DeltaLiftedDeltaLiftedJoin[ProvenanceIndexedRewrite, Signal, Fact]
    index: DeltaLiftedDeltaLiftedJoin[Fact, JoinIndex, IndexedFact]

    # Distincts
    distinct_rewrites: DeltaLiftedDeltaLiftedDistinct[ProvenanceIndexedRewrite]
    distinct_facts: DeltaLiftedDeltaLiftedDistinct[Fact]
    distinct_indexed_facts: DeltaLiftedDeltaLiftedDistinct[IndexedFact]

    # Delays
    delay_distinct_indexed_facts: Delay[Stream[ZSet[IndexedFact]]]
    delay_distinct_rewrites: Delay[Stream[ZSet[ProvenanceIndexedRewrite]]]

    # Pluses
    fresh_facts_plus_edb: LiftedGroupAdd[Stream[EDB]]
    rewrite_product_plus_rewrites: LiftedGroupAdd[Stream[ZSet[ProvenanceIndexedRewrite]]]

    # Stream elimination
    lift_elim_fresh_facts: LiftedStreamElimination[EDB]

    def set_input_a(self, stream_handle_a: StreamHandle[EDB]) -> None:
        self.input_stream_handle_a = stream_handle_a
        self.lift_intro_edb = LiftedStreamIntroduction(self.input_stream_handle_a)

        provenance_indexed_rewrite_group: ZSetAddition[ProvenanceIndexedRewrite] = ZSetAddition()
        rewrite_stream: Stream[ZSet[ProvenanceIndexedRewrite]] = Stream(provenance_indexed_rewrite_group)
        self.rewrites = StreamHandle(lambda: rewrite_stream)
        empty_rewrite_set = rewrite_stream.group().identity()
        empty_rewrite_set.inner[(0, RewriteMonoid().identity())] = 1

        self.rewrites.get().send(empty_rewrite_set)
        self.lift_rewrites = LiftedStreamIntroduction(self.rewrites)

    def set_input_b(self, stream_handle_b: StreamHandle[Program]) -> None:
        self.input_stream_handle_b = stream_handle_b

        self.lifted_sig = LiftedSig(self.input_stream_handle_b)
        self.lifted_intro_lifted_sig = LiftedStreamIntroduction(self.lifted_sig.output_handle())

        self.lifted_dir = LiftedDir(self.input_stream_handle_b)
        self.lifted_intro_lifted_dir = LiftedStreamIntroduction(self.lifted_dir.output_handle())

        self.lifted_jorder = LiftedJorder(self.input_stream_handle_b)
        self.lifted_intro_lifted_jorder = LiftedStreamIntroduction(self.lifted_jorder.output_handle())

        self.gatekeep = DeltaLiftedDeltaLiftedJoin(
            None,
            None,
            lambda left, right: left[0] == right[0],
            lambda left, right: (right[1], right[2], left[1]),
        )
        self.indexed_gatekeep = LiftedLiftedIndex(self.gatekeep.output_handle(), lambda s: s[0])

        self.product = DeltaLiftedDeltaLiftedJoin(
            self.gatekeep.output_handle(),
            None,
            lambda left, right: left[1] is None
            or (left[1][0] == right[1][0] and left[2] == RewriteMonoid().identity())
            or (left[1][0] == right[1][0] and (get_constant_terms(left[2].apply(left[1]))) == right[0][1]),
            lambda left, right: rewrite_product_projection((left[0], left[1], left[2]), right[1]),
        )

        self.ground = DeltaLiftedDeltaLiftedJoin(
            self.product.output_handle(),
            self.lifted_intro_lifted_sig.output_handle(),
            lambda left, right: left[0] == right[0],
            lambda left, right: left[1].apply(right[1]),
        )

        self.fresh_facts_plus_edb = LiftedGroupAdd(self.ground.output_handle(), self.lift_intro_edb.output_handle())
        self.distinct_facts = DeltaLiftedDeltaLiftedDistinct(self.fresh_facts_plus_edb.output_handle())
        self.index = DeltaLiftedDeltaLiftedJoin(
            self.distinct_facts.output_handle(),
            self.lifted_intro_lifted_jorder.output_handle(),
            lambda left, right: left[0] == right[0],
            lambda left, right: index_fact(right[1], left),
        )
        self.distinct_indexed_facts = DeltaLiftedDeltaLiftedDistinct(self.index.output_handle())

        self.rewrite_product_plus_rewrites = LiftedGroupAdd(
            self.product.output_handle(), self.lift_rewrites.output_handle()
        )
        self.distinct_rewrites = DeltaLiftedDeltaLiftedDistinct(self.rewrite_product_plus_rewrites.output_handle())

        self.delay_distinct_rewrites = Delay(self.distinct_rewrites.output_handle())
        self.delay_distinct_indexed_facts = Delay(self.distinct_indexed_facts.output_handle())

        self.gatekeep.set_input_a(self.delay_distinct_rewrites.output_handle())
        self.gatekeep.set_input_b(self.lifted_intro_lifted_dir.output_handle())
        self.product.set_input_b(self.delay_distinct_indexed_facts.output_handle())

        self.lift_elim_fresh_facts = LiftedStreamElimination(self.distinct_facts.output_handle())
        self.output_stream_handle = self.lift_elim_fresh_facts.output_handle()

    def step(self) -> bool:
        self.lift_intro_edb.step()
        self.lift_rewrites.step()
        self.lifted_sig.step()
        self.lifted_dir.step()
        self.lifted_jorder.step()
        self.lifted_intro_lifted_sig.step()
        self.lifted_intro_lifted_dir.step()
        self.lifted_intro_lifted_jorder.step()
        self.gatekeep.step()
        self.product.step()
        self.ground.step()
        self.fresh_facts_plus_edb.step()
        self.distinct_facts.step()
        self.index.step()
        self.distinct_indexed_facts.step()
        self.rewrite_product_plus_rewrites.step()
        self.distinct_rewrites.step()
        self.delay_distinct_indexed_facts.step()
        self.delay_distinct_rewrites.step()
        self.lift_elim_fresh_facts.step()

        new_facts_stream = self.distinct_facts.output().latest()
        new_rewrites_stream = self.distinct_rewrites.output().latest()
        return new_facts_stream.is_identity() and new_rewrites_stream.is_identity()



from typing import Optional, TypeVar

from pydbsp.core import AbelianGroupOperation
from pydbsp.stream import F2, BinaryOperator, Lift2, StreamHandle
from pydbsp.stream.operators.linear import Delay, Integrate

T = TypeVar("T")
R = TypeVar("R")
S = TypeVar("S")


class Incrementalize2(BinaryOperator[T, R, S]):
    """
    Given some bilinear function f, it both lifts it AND makes it incremental.

    For instance, if f is the relational join ⨝ over Z-sets **a** and **b**, incrementalizing it
    would create a circuit with three joins, yielding at each timestamp t: a[t] ⨝ b[t] + z^1(I(a))[t] ⨝ b[t] + a[t] ⨝ z^1(I(b))[t]

    Tl;dr this makes the join react to streams of additions and deletions

    """

    integrated_stream_a: Integrate[T]
    delayed_integrated_stream_a: Delay[T]

    integrated_stream_b: Integrate[R]
    delayed_integrated_stream_b: Delay[R]

    f2_a_b: Lift2[T, R, S]
    f2_delayed_integrated_a_b: Lift2[T, R, S]
    f2_a_delayed_integrated_b: Lift2[T, R, S]

    def __init__(
        self,
        stream_a: Optional[StreamHandle[T]],
        stream_b: Optional[StreamHandle[R]],
        f2: F2[T, R, S],
        output_stream_group: Optional[AbelianGroupOperation[S]],
    ) -> None:
        self.f2 = f2
        super().__init__(stream_a, stream_b, output_stream_group)

    def set_input_a(self, stream_handle_a: StreamHandle[T]) -> None:
        self.input_stream_handle_a = stream_handle_a
        self.integrated_stream_a = Integrate(self.input_stream_handle_a)
        self.delayed_integrated_stream_a = Delay(self.integrated_stream_a.output_handle())

    def set_input_b(self, stream_handle_b: StreamHandle[R]) -> None:
        self.input_stream_handle_b = stream_handle_b
        self.integrated_stream_b = Integrate(self.input_stream_handle_b)
        self.delayed_integrated_stream_b = Delay(self.integrated_stream_b.output_handle())

        self.f2_a_b = Lift2(self.input_stream_handle_a, self.input_stream_handle_b, self.f2, None)
        self.f2_a_delayed_integrated_b = Lift2(
            self.input_stream_handle_a, self.delayed_integrated_stream_b.output_handle(), self.f2, None
        )
        self.f2_delayed_integrated_a_b = Lift2(
            self.delayed_integrated_stream_a.output_handle(), self.input_stream_handle_b, self.f2, None
        )

    def step(self) -> bool:
        """Computes a[t] ⨝ b[t] + z^1(I(a))[t] ⨝ b[t] + a[t] ⨝ z^1(I(b))[t]"""
        self.integrated_stream_a.step()
        self.delayed_integrated_stream_a.step()

        self.integrated_stream_b.step()
        self.delayed_integrated_stream_b.step()

        self.f2_a_b.step()
        self.f2_a_delayed_integrated_b.step()
        self.f2_delayed_integrated_a_b.step()

        ab = self.f2_a_b.output().latest()
        adib = self.f2_a_delayed_integrated_b.output().latest()
        diab = self.f2_delayed_integrated_a_b.output().latest()

        group = self.output().group()
        self.output().send(group.add(group.add(ab, adib), diab))

        return True



from typing import Optional, TypeVar

from pydbsp.stream import Lift1, Stream, StreamHandle, step_until_fixpoint_and_return
from pydbsp.zset import ZSet
from pydbsp.zset.functions.linear import Cmp, Projection, project, select

T = TypeVar("T")


class LiftedSelect(Lift1[ZSet[T], ZSet[T]]):
    def __init__(self, stream: Optional[StreamHandle[ZSet[T]]], p: Cmp[T]):
        super().__init__(stream, lambda z: select(z, p), None)


class LiftedLiftedSelect(Lift1[Stream[ZSet[T]], Stream[ZSet[T]]]):
    def __init__(self, stream: Optional[StreamHandle[Stream[ZSet[T]]]], p: Cmp[T]):
        super().__init__(
            stream,
            lambda x: step_until_fixpoint_and_return(LiftedSelect(StreamHandle(lambda: x), p)),
            None,
        )


R = TypeVar("R")


class LiftedProject(Lift1[ZSet[T], ZSet[R]]):
    def __init__(self, stream: Optional[StreamHandle[ZSet[T]]], f: Projection[T, R]):
        super().__init__(stream, lambda z: project(z, f), None)


class LiftedLiftedProject(Lift1[Stream[ZSet[T]], Stream[ZSet[R]]]):
    def __init__(self, stream: Optional[StreamHandle[Stream[ZSet[T]]]], f: Projection[T, R]):
        super().__init__(
            stream,
            lambda x: step_until_fixpoint_and_return(LiftedProject(StreamHandle(lambda: x), f)),
            None,
        )



from typing import TypeVar

from pydbsp.stream import (
    Lift2,
    Stream,
    StreamHandle,
    step_until_fixpoint_and_return,
)
from pydbsp.zset import ZSet
from pydbsp.zset.functions.binary import H

T = TypeVar("T")


class LiftedH(Lift2[ZSet[T], ZSet[T], ZSet[T]]):
    def __init__(
        self,
        diff_stream_a: StreamHandle[ZSet[T]],
        integrated_stream_a: StreamHandle[ZSet[T]],
    ):
        super().__init__(diff_stream_a, integrated_stream_a, H, None)


class LiftedLiftedH(Lift2[Stream[ZSet[T]], Stream[ZSet[T]], Stream[ZSet[T]]]):
    def __init__(
        self,
        integrated_diff_stream_a: StreamHandle[Stream[ZSet[T]]],
        lifted_delayed_lifted_integrated_stream_a: StreamHandle[Stream[ZSet[T]]],
    ):
        super().__init__(
            integrated_diff_stream_a,
            lifted_delayed_lifted_integrated_stream_a,
            lambda x, y: step_until_fixpoint_and_return(LiftedH(StreamHandle(lambda: x), StreamHandle(lambda: y))),
            None,
        )



import sys
from abc import abstractmethod
from types import NotImplementedType
from typing import Callable, Generic, Iterator, List, Optional, Protocol, TypeVar, cast

from pydbsp.core import AbelianGroupOperation

T = TypeVar("T")

INFINITY = sys.maxsize


class Stream(Generic[T]):
    """
    Represents a stream of elements from an Abelian group.
    """

    timestamp: int
    inner: List[T]
    group_op: AbelianGroupOperation[T]
    identity: bool

    def __init__(self, group_op: AbelianGroupOperation[T]) -> None:
        self.inner = []
        self.group_op = group_op
        self.timestamp = -1
        self.identity = True
        self.send(group_op.identity())

    def send(self, element: T) -> None:
        """Adds an element to the stream and increments the timestamp."""
        id = self.group().identity()
        if element != id:
            self.identity = False

        self.inner.append(element)
        self.timestamp += 1

    def group(self) -> AbelianGroupOperation[T]:
        """Returns the Abelian group operation associated with this stream."""
        return self.group_op

    def current_time(self) -> int:
        """Returns the timestamp of the most recently arrived element."""
        return self.timestamp

    def __iter__(self) -> Iterator[T]:
        return self.inner.__iter__()

    def __repr__(self) -> str:
        return self.inner.__repr__()

    def __getitem__(self, timestamp: int) -> T:
        """Returns the element at the given timestamp."""
        if timestamp < 0:
            raise ValueError("Timestamp cannot be negative")

        if timestamp <= self.current_time():
            return self.inner.__getitem__(timestamp)

        elif timestamp > self.current_time():
            id = self.group().identity()
            while timestamp > self.current_time():
                self.send(id)

        return self.__getitem__(timestamp)

    def latest(self) -> T:
        """Returns the most recent element."""
        return self.__getitem__(self.current_time())

    def is_identity(self) -> bool:
        return self.identity

    def __eq__(self, other: object) -> bool | NotImplementedType:
        """
        Compares this stream with another, considering all timestamps up to the latest.
        """
        if not isinstance(other, Stream):
            return NotImplemented

        if self.is_identity() and other.is_identity():
            return True

        cast(Stream[T], other)

        self_timestamp = self.current_time()
        other_timestamp = other.current_time()

        if self_timestamp != other_timestamp:
            return False

        return self.inner == other.inner  # type: ignore


StreamReference = Callable[[], Stream[T]]


class StreamHandle(Generic[T]):
    """A handle to a stream, allowing lazy access."""

    ref: StreamReference[T]

    def __init__(self, stream_reference: StreamReference[T]) -> None:
        self.ref = stream_reference

    def get(self) -> Stream[T]:
        """Returns the referenced stream."""
        return self.ref()


R = TypeVar("R")


class Operator(Protocol[T]):
    @abstractmethod
    def step(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def output_handle(self) -> StreamHandle[T]:
        raise NotImplementedError


def step_until_fixpoint[T](operator: Operator[T]) -> None:
    while not operator.step():
        pass


def step_until_fixpoint_and_return[T](operator: Operator[T]) -> Stream[T]:
    step_until_fixpoint(operator)

    return operator.output_handle().get()


class UnaryOperator(Operator[R], Protocol[T, R]):
    """Base class for stream operators with a single input and output."""

    input_stream_handle: StreamHandle[T]
    output_stream_handle: StreamHandle[R]

    def __init__(
        self,
        stream_handle: Optional[StreamHandle[T]],
        output_stream_group: Optional[AbelianGroupOperation[R]],
    ) -> None:
        if stream_handle is not None:
            self.set_input(stream_handle, output_stream_group)

    def set_input(
        self,
        stream_handle: StreamHandle[T],
        output_stream_group: Optional[AbelianGroupOperation[R]],
    ) -> None:
        """Sets the input stream and initializes the output stream."""
        self.input_stream_handle = stream_handle

        if output_stream_group is not None:
            output = Stream(output_stream_group)

            self.output_stream_handle = StreamHandle(lambda: output)
        else:
            output = cast(Stream[R], Stream(self.input_a().group()))

            self.output_stream_handle = StreamHandle(lambda: output)

    def output(self) -> Stream[R]:
        return self.output_stream_handle.get()

    def input_a(self) -> Stream[T]:
        return self.input_stream_handle.get()

    def output_handle(self) -> StreamHandle[R]:
        handle = StreamHandle(lambda: self.output())

        return handle


S = TypeVar("S")


class BinaryOperator(Operator[S], Protocol[T, R, S]):
    """Base class for stream operators with two inputs and one output."""

    input_stream_handle_a: StreamHandle[T]
    input_stream_handle_b: StreamHandle[R]
    output_stream_handle: StreamHandle[S]

    def __init__(
        self,
        stream_a: Optional[StreamHandle[T]],
        stream_b: Optional[StreamHandle[R]],
        output_stream_group: Optional[AbelianGroupOperation[S]],
    ) -> None:
        if stream_a is not None:
            self.set_input_a(stream_a)

        if stream_b is not None:
            self.set_input_b(stream_b)

        if output_stream_group is not None:
            output = Stream(output_stream_group)

            self.set_output_stream(StreamHandle(lambda: output))

    def set_input_a(self, stream_handle_a: StreamHandle[T]) -> None:
        """Sets the first input stream and initializes the output stream."""
        self.input_stream_handle_a = stream_handle_a
        output = cast(Stream[S], Stream(self.input_a().group()))

        self.set_output_stream(StreamHandle(lambda: output))

    def set_input_b(self, stream_handle_b: StreamHandle[R]) -> None:
        """Sets the second input stream."""
        self.input_stream_handle_b = stream_handle_b

    def set_output_stream(self, output_stream_handle: StreamHandle[S]) -> None:
        """Sets the output stream handle."""
        self.output_stream_handle = output_stream_handle

    def output(self) -> Stream[S]:
        return self.output_stream_handle.get()

    def input_a(self) -> Stream[T]:
        return self.input_stream_handle_a.get()

    def input_b(self) -> Stream[R]:
        return self.input_stream_handle_b.get()

    def output_handle(self) -> StreamHandle[S]:
        handle = StreamHandle(lambda: self.output())

        return handle


F1 = Callable[[T], R]


class Lift1(UnaryOperator[T, R]):
    """Lifts a unary function to operate on a stream"""

    f1: F1[T, R]
    frontier: int

    def __init__(
        self,
        stream: Optional[StreamHandle[T]],
        f1: F1[T, R],
        output_stream_group: Optional[AbelianGroupOperation[R]],
    ):
        self.f1 = f1
        self.frontier = 0
        super().__init__(stream, output_stream_group)

    def step(self) -> bool:
        """Applies the lifted function to the next element in the input stream."""
        output_timestamp = self.output().current_time()
        input_timestamp = self.input_a().current_time()
        join = max(input_timestamp, output_timestamp, self.frontier)
        meet = min(input_timestamp, output_timestamp, self.frontier)

        if join == meet:
            return True

        next_frontier = self.frontier + 1
        self.output().send(self.f1(self.input_a()[next_frontier]))
        self.frontier = next_frontier

        return False


F2 = Callable[[T, R], S]


class Lift2(BinaryOperator[T, R, S]):
    """Lifts a binary function to operate on two streams"""

    f2: F2[T, R, S]
    frontier_a: int
    frontier_b: int

    def __init__(
        self,
        stream_a: Optional[StreamHandle[T]],
        stream_b: Optional[StreamHandle[R]],
        f2: F2[T, R, S],
        output_stream_group: Optional[AbelianGroupOperation[S]],
    ) -> None:
        self.f2 = f2
        self.frontier_a = 0
        self.frontier_b = 0

        super().__init__(stream_a, stream_b, output_stream_group)

    def step(self) -> bool:
        """Applies the lifted function to the most recently arrived elements in both input streams."""
        a_timestamp = self.input_a().current_time()
        b_timestamp = self.input_b().current_time()
        output_timestamp = self.output().current_time()

        join = max(a_timestamp, b_timestamp, output_timestamp, self.frontier_a, self.frontier_b)
        meet = min(a_timestamp, b_timestamp, output_timestamp, self.frontier_a, self.frontier_b)
        if join == meet:
            return True

        next_frontier_a = self.frontier_a + 1
        next_frontier_b = self.frontier_b + 1
        a = self.input_a()[next_frontier_a]
        b = self.input_b()[next_frontier_b]

        application = self.f2(a, b)
        self.output().send(application)

        self.frontier_a = next_frontier_a
        self.frontier_b = next_frontier_b

        return False


class LiftedGroupAdd(Lift2[T, T, T]):
    def __init__(self, stream_a: StreamHandle[T], stream_b: Optional[StreamHandle[T]]):
        super().__init__(
            stream_a,
            stream_b,
            lambda x, y: stream_a.get().group().add(x, y),
            None,
        )


class LiftedGroupNegate(Lift1[T, T]):
    def __init__(self, stream: StreamHandle[T]):
        super().__init__(stream, lambda x: stream.get().group().neg(x), None)


class StreamAddition(AbelianGroupOperation[Stream[T]]):
    """Defines addition for streams by lifting their underlying group's addition."""

    group: AbelianGroupOperation[T]

    def __init__(self, group: AbelianGroupOperation[T]) -> None:
        self.group = group

    def add(self, a: Stream[T], b: Stream[T]) -> Stream[T]:
        """Adds two streams element-wise."""
        handle_a = StreamHandle(lambda: a)
        handle_b = StreamHandle(lambda: b)

        lifted_group_add = LiftedGroupAdd(handle_a, handle_b)

        return step_until_fixpoint_and_return(lifted_group_add)

    def inner_group(self) -> AbelianGroupOperation[T]:
        """Returns the underlying group operation."""
        return self.group

    def neg(self, a: Stream[T]) -> Stream[T]:
        """Negates a stream element-wise."""
        handle_a = StreamHandle(lambda: a)
        lifted_group_neg = LiftedGroupNegate(handle_a)

        return step_until_fixpoint_and_return(lifted_group_neg)

    def identity(self) -> Stream[T]:
        """
        Returns an identity stream for the addition operation.
        """
        identity_stream = Stream(self.group)

        return identity_stream



from typing import Optional

from pydbsp.core import AbelianGroupOperation
from pydbsp.indexed_zset import IndexedZSet, IndexedZSetAddition, Indexer
from pydbsp.indexed_zset.functions.linear import index_zset
from pydbsp.stream import Lift1, Stream, StreamHandle, step_until_fixpoint_and_return
from pydbsp.zset import ZSet, ZSetAddition


class LiftedIndex[I, T](Lift1[ZSet[T], IndexedZSet[I, T]]):
    """Creates a stream where the output at each timestamp is the input ZSet indexed according to some indexing function"""

    indexer: Indexer[T, I]

    def __init__(self, stream: Optional[StreamHandle[ZSet[T]]], indexer: Indexer[T, I]):
        self.indexer = indexer
        super().__init__(stream, lambda z: index_zset(z, indexer), None)

    def set_input(
        self,
        stream_handle: StreamHandle[ZSet[T]],
        output_stream_group: Optional[AbelianGroupOperation[IndexedZSet[I, T]]],
    ) -> None:
        self.input_stream_handle = stream_handle
        inner_group: ZSetAddition[T] = ZSetAddition()
        group = IndexedZSetAddition(inner_group, self.indexer)
        output = Stream(group)

        self.output_stream_handle = StreamHandle(lambda: output)


class LiftedLiftedIndex[I, T](Lift1[Stream[ZSet[T]], Stream[IndexedZSet[I, T]]]):
    """Creates a stream where the output at each timestamp is the input ZSet indexed according to some indexing function"""

    def __init__(self, stream: Optional[StreamHandle[Stream[ZSet[T]]]], indexer: Indexer[T, I]):
        super().__init__(
            stream, lambda sp: step_until_fixpoint_and_return(LiftedIndex(StreamHandle(lambda: sp), indexer)), None
        )



from pydbsp.lazy_zset import LazyZSet
from pydbsp.zset.functions.linear import Cmp, Projection
from pydbsp.zset.functions.linear import project as zset_project
from pydbsp.zset.functions.linear import select as zset_select


def select[T](lazy_zset: LazyZSet[T], p: Cmp[T]) -> LazyZSet[T]:
    """Filters the given Lazy Z-set based on a predicate function."""
    return LazyZSet([zset_select(zset, p) for zset in lazy_zset])


def project[T, R](lazy_zset: LazyZSet[T], f: Projection[T, R]) -> LazyZSet[R]:
    """
    Projects a Lazy Z-set unto a new Lazy Z-set by applying a function to each element.

    Returns a new Lazy Z-set where each element is the result of applying f to an element
    of the input Lazy Z-set. Weights of elements mapping to the same value are summed.
    """
    return LazyZSet([zset_project(zset, f) for zset in lazy_zset])



from typing import Dict, Generic, Iterable, Tuple, TypeVar

from pydbsp.core import AbelianGroupOperation

T = TypeVar("T")


class ZSet(Generic[T]):
    """
    Represents a Z-set, a generalization of multisets with integer weights.
    Elements can have positive, negative, or zero weights.

    A Z-Set whose elements have all weight one can be interpreted as a set. One where
    all are strictly positive is a bag, and one where they are either one or -1 is a diff.
    """

    inner: Dict[T, int]

    def __init__(self, values: Dict[T, int]) -> None:
        self.inner = values

    def items(self) -> Iterable[Tuple[T, int]]:
        """Returns an iterable of (element, weight) pairs."""
        return self.inner.items()

    def __repr__(self) -> str:
        return self.inner.__repr__()

    def __eq__(self, other: object) -> bool:
        """
        Two Z-sets are equal if they have the same elements with the same weight.
        """
        if not isinstance(other, ZSet):
            return False

        return self.inner == other.inner  # type: ignore

    def __contains__(self, item: T) -> bool:
        """An item is in the Z-set if it has non-zero weight."""
        return self.inner.__contains__(item)

    def __getitem__(self, item: T) -> int:
        """Returns the weight of an item (0 if not present)."""
        if item not in self:
            return 0

        return self.inner[item]

    def is_identity(self) -> bool:
        return len(self.inner) == 0

    def __setitem__(self, key: T, value: int) -> None:
        self.inner[key] = value


class ZSetAddition(Generic[T], AbelianGroupOperation[ZSet[T]]):
    """
    Defines addition operation for Z-sets, forming an Abelian group.
    """

    def add(self, a: ZSet[T], b: ZSet[T]) -> ZSet[T]:
        """
        Adds two Z-sets by summing weights of common elements.
        Elements with resulting zero weight are removed.
        """
        c = {k: v for k, v in a.inner.items() if v != 0}
        for k, v in b.inner.items():
            if k in c:
                new_weight = c[k] + v
                if new_weight != 0:
                    c[k] = new_weight
                else:
                    del c[k]
            else:
                c[k] = v

        return ZSet(c)

    def neg(self, a: ZSet[T]) -> ZSet[T]:
        """Returns the inverse of a Z-set by negating all weights."""
        return ZSet({k: v * -1 for k, v in a.inner.items()})

    def identity(self) -> ZSet[T]:
        """Returns the empty Z-set."""
        return ZSet({})



from pydbsp.core import AbelianGroupOperation
from pydbsp.stream import Stream


def stream_introduction[T](value: T, group: AbelianGroupOperation[T]) -> Stream[T]:
    """
    Creates a stream out of a single element and an associated abelian group.

    Args:
        value (T): The value to be introduced into the stream.
        group (AbelianGroupOperation[T]): The group operation for the stream.

    Returns:
        Stream[T]: A new stream containing the single value.
    """
    output_stream = Stream(group)
    output_stream.send(value)

    return output_stream


def stream_elimination[T](stream: Stream[T]) -> T:
    """
    Sums all elements in a stream using the stream's group operation.

    Args:
        stream (Stream[T]): The input stream to be "squashed".

    Returns:
        T: The result of combining all elements in the stream.
    """
    output_value = stream.group().identity()
    for value in stream:
        output_value = stream.group().add(output_value, value)

    return output_value



from typing import Dict

from pydbsp.lazy_zset import LazyZSet
from pydbsp.zset import ZSet


def H[T](diff: LazyZSet[T], integrated_state: LazyZSet[T]) -> LazyZSet[T]:
    distincted_diff: Dict[T, int] = {}

    jiff = diff.coalesce()
    # for diff_zset in jiff:
    for k, v in jiff.items():
        current_k_latest_diff_weight = v

        if k in integrated_state:
            current_k_latest_delayed_state_weight = integrated_state[k]
            coalesced_weight = current_k_latest_diff_weight + current_k_latest_delayed_state_weight

            if current_k_latest_delayed_state_weight > 0 and coalesced_weight <= 0:
                distincted_diff[k] = -1

                continue

            if current_k_latest_delayed_state_weight <= 0 and coalesced_weight > 0:
                distincted_diff[k] = 1
        else:
            if v > 0:
                distincted_diff[k] = 1

    return LazyZSet([ZSet(distincted_diff)])



from typing import Optional

from pydbsp.lazy_zset import LazyZSet, LazyZSetAddition
from pydbsp.lazy_zset.functions.bilinear import join
from pydbsp.stream import (
    BinaryOperator,
    Lift2,
    LiftedGroupAdd,
    Stream,
    StreamAddition,
    StreamHandle,
    step_until_fixpoint_and_return,
)
from pydbsp.stream.operators.linear import Delay, Integrate, LiftedDelay, LiftedIntegrate
from pydbsp.zset.functions.bilinear import JoinCmp, PostJoinProjection


class LiftedJoin[T, R, S](Lift2[LazyZSet[T], LazyZSet[R], LazyZSet[S]]):
    def __init__(
        self,
        stream_a: Optional[StreamHandle[LazyZSet[T]]],
        stream_b: Optional[StreamHandle[LazyZSet[R]]],
        p: JoinCmp[T, R],
        f: PostJoinProjection[T, R, S],
    ):
        super().__init__(stream_a, stream_b, lambda x, y: join(x, y, p, f), None)


class LiftedLiftedJoin[T, R, S](
    Lift2[
        Stream[LazyZSet[T]],
        Stream[LazyZSet[R]],
        Stream[LazyZSet[S]],
    ]
):
    """
    Computes the Lazy Z-Set join between two streams element-wise.
    """

    def __init__(
        self,
        stream_a: Optional[StreamHandle[Stream[LazyZSet[T]]]],
        stream_b: Optional[StreamHandle[Stream[LazyZSet[R]]]],
        p: JoinCmp[T, R],
        f: PostJoinProjection[T, R, S],
    ):
        super().__init__(
            stream_a,
            stream_b,
            lambda x, y: step_until_fixpoint_and_return(
                LiftedJoin(StreamHandle(lambda: x), StreamHandle(lambda: y), p, f)
            ),
            None,
        )


class DeltaLiftedDeltaLiftedJoin[T, R, S](
    BinaryOperator[Stream[LazyZSet[T]], Stream[LazyZSet[R]], Stream[LazyZSet[S]]]
):
    """
    Incrementally computes the Lazy Z-Set join between two streams element-wise. Equivalent to - but keeps less state - incrementalizing a doubly-lifted join. See :func:`~pydbsp.stream.operators.Incrementalize2` to grasp what it means to incrementalize a singly-lifted join.
    """

    p: JoinCmp[T, R]
    f: PostJoinProjection[T, R, S]
    frontier_a: int
    frontier_b: int

    integrated_stream_a: Integrate[Stream[LazyZSet[T]]]
    delayed_integrated_stream_a: Delay[Stream[LazyZSet[T]]]
    lift_integrated_stream_a: LiftedIntegrate[LazyZSet[T]]
    integrated_lift_integrated_stream_a: Integrate[Stream[LazyZSet[T]]]

    integrated_stream_b: Integrate[Stream[LazyZSet[R]]]
    delayed_integrated_stream_b: Delay[Stream[LazyZSet[R]]]
    lift_integrated_stream_b: LiftedIntegrate[LazyZSet[R]]
    integrated_lift_integrated_stream_b: Integrate[Stream[LazyZSet[R]]]
    lift_delayed_integrated_lift_integrated_stream_b: LiftedDelay[LazyZSet[R]]
    lift_delayed_lift_integrated_stream_b: LiftedDelay[LazyZSet[R]]

    join_1: LiftedLiftedJoin[T, R, S]
    join_2: LiftedLiftedJoin[T, R, S]
    join_3: LiftedLiftedJoin[T, R, S]
    join_4: LiftedLiftedJoin[T, R, S]

    sum_one: LiftedGroupAdd[Stream[LazyZSet[S]]]
    sum_two: LiftedGroupAdd[Stream[LazyZSet[S]]]
    sum_three: LiftedGroupAdd[Stream[LazyZSet[S]]]

    output_stream: Stream[Stream[LazyZSet[S]]]

    def set_input_a(self, stream_handle_a: StreamHandle[Stream[LazyZSet[T]]]) -> None:
        self.input_stream_handle_a = stream_handle_a
        self.integrated_stream_a = Integrate(self.input_stream_handle_a)
        self.delayed_integrated_stream_a = Delay(self.integrated_stream_a.output_handle())

        self.lift_integrated_stream_a = LiftedIntegrate(self.input_stream_handle_a)
        self.integrated_lift_integrated_stream_a = Integrate(self.lift_integrated_stream_a.output_handle())

    def set_input_b(self, stream_handle_b: StreamHandle[Stream[LazyZSet[R]]]) -> None:
        self.input_stream_handle_b = stream_handle_b
        self.integrated_stream_b = Integrate(self.input_stream_handle_b)
        self.delayed_integrated_stream_b = Delay(self.integrated_stream_b.output_handle())

        self.lift_integrated_stream_b = LiftedIntegrate(self.input_stream_handle_b)
        self.integrated_lift_integrated_stream_b = Integrate(self.lift_integrated_stream_b.output_handle())
        self.lift_delayed_integrated_lift_integrated_stream_b = LiftedDelay(
            self.integrated_lift_integrated_stream_b.output_handle()
        )
        self.lift_delayed_lift_integrated_stream_b = LiftedDelay(self.lift_integrated_stream_b.output_handle())

        self.join_1 = LiftedLiftedJoin(
            self.delayed_integrated_stream_a.output_handle(),
            self.lift_delayed_lift_integrated_stream_b.output_handle(),
            self.p,
            self.f,
        )
        self.join_2 = LiftedLiftedJoin(
            self.integrated_lift_integrated_stream_a.output_handle(),
            self.input_stream_handle_b,
            self.p,
            self.f,
        )
        self.join_3 = LiftedLiftedJoin(
            self.lift_integrated_stream_a.output_handle(),
            self.delayed_integrated_stream_b.output_handle(),
            self.p,
            self.f,
        )
        self.join_4 = LiftedLiftedJoin(
            self.input_stream_handle_a,
            self.lift_delayed_integrated_lift_integrated_stream_b.output_handle(),
            self.p,
            self.f,
        )

        self.sum_one = LiftedGroupAdd(self.join_1.output_handle(), self.join_2.output_handle())
        self.sum_two = LiftedGroupAdd(self.sum_one.output_handle(), self.join_3.output_handle())
        self.sum_three = LiftedGroupAdd(self.sum_two.output_handle(), self.join_4.output_handle())

    def __init__(
        self,
        diff_stream_a: Optional[StreamHandle[Stream[LazyZSet[T]]]],
        diff_stream_b: Optional[StreamHandle[Stream[LazyZSet[R]]]],
        p: JoinCmp[T, R],
        f: PostJoinProjection[T, R, S],
    ):
        self.p = p
        self.f = f
        self.frontier_a = 0
        self.frontier_b = 0
        inner_group: LazyZSetAddition[S] = LazyZSetAddition()
        group: StreamAddition[LazyZSet[S]] = StreamAddition(inner_group)  # type: ignore

        self.output_stream = Stream(group)
        self.output_stream_handle = StreamHandle(lambda: self.output_stream)

        if diff_stream_a is not None:
            self.set_input_a(diff_stream_a)

        if diff_stream_b is not None:
            self.set_input_b(diff_stream_b)

    def output(self) -> Stream[Stream[LazyZSet[S]]]:
        return self.output_stream

    def step(self) -> bool:
        current_a_timestamp = self.input_a().current_time()
        current_b_timestamp = self.input_b().current_time()
        # Not sure about this.
        if current_a_timestamp == self.frontier_a and current_b_timestamp == self.frontier_b:
            return True

        self.integrated_stream_a.step()
        self.delayed_integrated_stream_a.step()
        self.lift_integrated_stream_a.step()
        self.integrated_lift_integrated_stream_a.step()
        self.integrated_stream_b.step()
        self.delayed_integrated_stream_b.step()
        self.lift_integrated_stream_b.step()
        self.integrated_lift_integrated_stream_b.step()
        self.lift_delayed_integrated_lift_integrated_stream_b.step()
        self.lift_delayed_lift_integrated_stream_b.step()
        self.join_1.step()
        self.join_2.step()
        self.join_3.step()
        self.join_4.step()

        group = self.output().group()
        sum_1 = group.add(self.join_1.output().latest(), self.join_2.output().latest())
        sum_2 = group.add(self.join_3.output().latest(), self.join_4.output().latest())
        self.output_stream.send(group.add(sum_1, sum_2))

        self.frontier_a += 1
        self.frontier_b += 1

        return False



from pydbsp.lazy_zset import LazyZSet
from pydbsp.zset.functions.bilinear import JoinCmp, PostJoinProjection
from pydbsp.zset.functions.bilinear import join as zset_join


def join[T, R, S](
    left_zset: LazyZSet[T],
    right_zset: LazyZSet[R],
    p: JoinCmp[T, R],
    f: PostJoinProjection[T, R, S],
) -> LazyZSet[S]:
    zset_product = [(left_df, right_df) for left_df in left_zset.inner for right_df in right_zset.inner]

    result = [zset_join(left_zset, right_zset, p, f) for left_zset, right_zset in zset_product]

    return LazyZSet(result)



from typing import Tuple

from pydbsp.lazy_zset import LazyZSet
from pydbsp.lazy_zset.operators.bilinear import DeltaLiftedDeltaLiftedJoin as LazyZSetDeltaLiftedDeltaLiftedJoin
from pydbsp.lazy_zset.operators.unary import DeltaLiftedDeltaLiftedDistinct as LazyZSetDeltaLiftedDeltaLiftedDistinct
from pydbsp.stream import LiftedGroupAdd, Stream, StreamHandle, UnaryOperator
from pydbsp.stream.operators.linear import LiftedDelay, LiftedStreamElimination, LiftedStreamIntroduction
from pydbsp.zset import ZSet
from pydbsp.zset.operators.bilinear import DeltaLiftedDeltaLiftedJoin as ZSetDeltaLiftedDeltaLiftedJoin
from pydbsp.zset.operators.unary import DeltaLiftedDeltaLiftedDistinct as ZSetDeltaLiftedDeltaLiftedDistinct

Edge = Tuple[int, int]
GraphZSet = ZSet[Edge]


class IncrementalGraphReachability(UnaryOperator[GraphZSet, GraphZSet]):
    delta_input: LiftedStreamIntroduction[GraphZSet]
    join: ZSetDeltaLiftedDeltaLiftedJoin[Edge, Edge, Edge]
    delta_input_join_sum: LiftedGroupAdd[Stream[GraphZSet]]
    distinct: ZSetDeltaLiftedDeltaLiftedDistinct[Edge]
    lift_delayed_distinct: LiftedDelay[GraphZSet]
    flattened_output: LiftedStreamElimination[GraphZSet]

    def __init__(self, stream: StreamHandle[GraphZSet]):
        self.input_stream_handle = stream

        self.delta_input = LiftedStreamIntroduction(self.input_stream_handle)

        self.join = ZSetDeltaLiftedDeltaLiftedJoin(
            None,
            None,
            lambda left, right: left[1] == right[0],
            lambda left, right: (left[0], right[1]),
        )
        self.delta_input_join_sum = LiftedGroupAdd(self.delta_input.output_handle(), self.join.output_handle())

        self.distinct = ZSetDeltaLiftedDeltaLiftedDistinct(self.delta_input_join_sum.output_handle())
        self.lift_delayed_distinct = LiftedDelay(self.distinct.output_handle())

        self.join.set_input_a(self.lift_delayed_distinct.output_handle())
        self.join.set_input_b(self.delta_input.output_handle())

        self.flattened_output = LiftedStreamElimination(self.distinct.output_handle())
        self.output_stream_handle = self.flattened_output.output_handle()

    def step(self) -> bool:
        self.delta_input.step()
        self.delta_input_join_sum.step()
        self.distinct.step()
        self.lift_delayed_distinct.step()
        self.join.step()
        self.flattened_output.step()

        latest = self.flattened_output.output().latest()
        id = self.output().group().identity()
        if latest == id:
            return True

        return False


LazyGraphZSet = LazyZSet[Edge]


class LazyIncrementalGraphReachability(UnaryOperator[LazyGraphZSet, LazyGraphZSet]):
    delta_input: LiftedStreamIntroduction[LazyGraphZSet]
    join: LazyZSetDeltaLiftedDeltaLiftedJoin[Edge, Edge, Edge]
    delta_input_join_sum: LiftedGroupAdd[Stream[LazyGraphZSet]]
    distinct: LazyZSetDeltaLiftedDeltaLiftedDistinct[Edge]
    lift_delayed_distinct: LiftedDelay[LazyGraphZSet]
    flattened_output: LiftedStreamElimination[LazyGraphZSet]

    def __init__(self, stream: StreamHandle[LazyGraphZSet]):
        self.input_stream_handle = stream

        self.delta_input = LiftedStreamIntroduction(self.input_stream_handle)

        self.join = LazyZSetDeltaLiftedDeltaLiftedJoin(
            None,
            None,
            lambda left, right: left[1] == right[0],
            lambda left, right: (left[0], right[1]),
        )
        self.delta_input_join_sum = LiftedGroupAdd(self.delta_input.output_handle(), self.join.output_handle())

        self.distinct = LazyZSetDeltaLiftedDeltaLiftedDistinct(self.delta_input_join_sum.output_handle())
        self.lift_delayed_distinct = LiftedDelay(self.distinct.output_handle())

        self.join.set_input_a(self.lift_delayed_distinct.output_handle())
        self.join.set_input_b(self.delta_input.output_handle())

        self.flattened_output = LiftedStreamElimination(self.distinct.output_handle())
        self.output_stream_handle = self.flattened_output.output_handle()

    def step(self) -> bool:
        self.delta_input.step()
        self.delta_input_join_sum.step()
        self.distinct.step()
        self.lift_delayed_distinct.step()
        self.join.step()
        self.flattened_output.step()

        latest = self.flattened_output.output().latest()
        id = self.output().group().identity()
        if latest == id:
            return True

        return False



from bisect import bisect_right, insort
from typing import Callable, Dict, Generator, Generic, Iterable, Iterator, List, Set, Tuple, TypeVar

from pydbsp.core import AbelianGroupOperation
from pydbsp.zset import ZSetAddition

T = TypeVar("T")


class AppendOnlySpine(Generic[T]):
    """
    A append-only flat B-Tree. Borrowed from `https://github.com/grantjenks/python-sortedcontainers` and `https://github.com/brurucy/indexset`. Only for internal use.
    """

    _len: int
    _load: int
    _lists: List[List[T]]
    _maxes: List[T]

    def __init__(self):
        self._len = 0
        self._load = 1024
        self._lists = []
        self._maxes = []

    def __len__(self) -> int:
        return self._len

    def add(self, value: T):
        _lists = self._lists
        _maxes = self._maxes

        if _maxes:
            pos = bisect_right(_maxes, value)  # type: ignore

            if pos == len(_maxes):
                pos -= 1
                _lists[pos].append(value)
                _maxes[pos] = value
            else:
                insort(_lists[pos], value)  # type: ignore

            self._expand(pos)
        else:
            _lists.append([value])
            _maxes.append(value)

        self._len += 1

    def __iter__(self) -> Generator[T, None, None]:
        for sublist in self._lists:
            yield from sublist

    def _expand(self, pos: int):
        _load = self._load
        _lists = self._lists

        if len(_lists[pos]) > (_load << 1):
            _maxes = self._maxes

            _lists_pos = _lists[pos]
            half = _lists_pos[_load:]
            del _lists_pos[_load:]
            _maxes[pos] = _lists_pos[-1]

            _lists.insert(pos + 1, half)
            _maxes.insert(pos + 1, half[-1])


def sort_merge_join[T](spine1: AppendOnlySpine[T], spine2: AppendOnlySpine[T]) -> Iterator[T]:
    iter1: Generator[T] = iter(spine1)
    iter2: Generator[T] = iter(spine2)

    try:
        item1 = next(iter1)
        item2 = next(iter2)

        while True:
            if item1 < item2:  # type: ignore
                item1 = next(iter1)
            elif item2 < item1:  # type: ignore
                item2 = next(iter2)
            else:
                yield item1
                item1 = next(iter1)
                item2 = next(iter2)

    except StopIteration:
        return


I = TypeVar("I")
Indexer = Callable[[T], I]


class IndexedZSet(Generic[I, T]):
    """
    Represents a Z-set, with a B-Tree index. See :func:`~pydbsp.zset.ZSet`.
    """

    inner: Dict[T, int]
    index_to_value: Dict[I, Set[T]]
    indexer: Callable[[T], I]
    index: AppendOnlySpine[I]

    def __init__(self, values: Dict[T, int], indexer: Indexer[T, I]) -> None:
        self.inner = values
        self.index_to_value = {}
        self.index = AppendOnlySpine()
        self.indexer = indexer

        for value in self.inner.keys():
            indexed_value = self.indexer(value)
            if indexed_value in self.index_to_value:
                self.index_to_value[indexed_value].add(value)
            else:
                self.index_to_value[indexed_value] = {value}
                self.index.add(indexed_value)

    def items(self) -> Iterable[Tuple[T, int]]:
        return self.inner.items()

    def __repr__(self) -> str:
        return self.inner.__repr__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, IndexedZSet):
            return False

        return self.inner == other.inner  # type: ignore

    def __contains__(self, item: T) -> bool:
        return self.inner.__contains__(item)

    def __getitem__(self, item: T) -> int:
        if item not in self:
            return 0

        return self.inner[item]

    def __setitem__(self, key: T, value: int) -> None:
        self.inner[key] = value
        indexed_value = self.indexer(key)
        if indexed_value in self.index_to_value:
            self.index_to_value[indexed_value].add(key)
        else:
            self.index_to_value[indexed_value] = {key}

        self.index.add(indexed_value)


class IndexedZSetAddition(Generic[I, T], AbelianGroupOperation[IndexedZSet[I, T]]):
    inner_group: ZSetAddition[T]
    indexer: Indexer[T, I]

    def __init__(self, inner_group: ZSetAddition[T], indexer: Indexer[T, I]) -> None:
        self.inner_group = inner_group
        self.indexer = indexer

    def add(self, a: IndexedZSet[I, T], b: IndexedZSet[I, T]) -> IndexedZSet[I, T]:
        if len(b.inner) == 0:
            return a

        c = {k: v for k, v in a.inner.items() if v != 0}
        for k, v in b.inner.items():
            if k in c:
                new_weight = c[k] + v
                if new_weight != 0:
                    c[k] = new_weight
                else:
                    del c[k]
            else:
                c[k] = v

        return IndexedZSet(c, self.indexer)

    def neg(self, a: IndexedZSet[I, T]) -> IndexedZSet[I, T]:
        empty_dict: Dict[T, int] = {}
        b = IndexedZSet(empty_dict, a.indexer)
        b.index = a.index
        b.index_to_value = a.index_to_value
        b.inner = {k: v * -1 for k, v in a.inner.items()}

        return b

    def identity(self) -> IndexedZSet[I, T]:
        empty_dict: Dict[T, int] = {}

        return IndexedZSet(empty_dict, self.indexer)



from typing import Optional, TypeVar

from pydbsp.stream import (
    Lift1,
    LiftedGroupAdd,
    LiftedGroupNegate,
    Stream,
    StreamAddition,
    StreamHandle,
    UnaryOperator,
    step_until_fixpoint_and_return,
)
from pydbsp.stream.functions.linear import stream_elimination, stream_introduction

T = TypeVar("T")


class Delay(UnaryOperator[T, T]):
    """
    Delays the input stream by one timestamp.
    """

    def __init__(self, stream: Optional[StreamHandle[T]]) -> None:
        super().__init__(stream, None)

    def step(self) -> bool:
        """
        Outputs the previous value from the input stream.
        """
        output_timestamp = self.output().current_time()
        input_timestamp = self.input_a().current_time()

        if output_timestamp <= input_timestamp:
            self.output().send(self.input_a()[output_timestamp])

            return False

        return True


class Differentiate(UnaryOperator[T, T]):
    """
    Computes the difference between consecutive elements in the input stream.
    """

    delayed_stream: Delay[T]
    delayed_negated_stream: LiftedGroupNegate[T]
    differentiation_stream: LiftedGroupAdd[T]

    def __init__(self, stream: StreamHandle[T]) -> None:
        self.input_stream_handle = stream
        self.delayed_stream = Delay(self.input_stream_handle)
        self.delayed_negated_stream = LiftedGroupNegate(self.delayed_stream.output_handle())
        self.differentiation_stream = LiftedGroupAdd(
            self.input_stream_handle, self.delayed_negated_stream.output_handle()
        )
        self.output_stream_handle = self.differentiation_stream.output_handle()

    def step(self) -> bool:
        """
        Outputs the difference between the latest element from the input stream with the one before
        """
        self.delayed_stream.step()
        self.delayed_negated_stream.step()
        self.differentiation_stream.step()

        return self.output().current_time() == self.input_a().current_time()


class Integrate(UnaryOperator[T, T]):
    """
    Computes the running sum of the input stream.
    """

    delayed_stream: Delay[T]
    integration_stream: LiftedGroupAdd[T]

    def __init__(self, stream: StreamHandle[T]) -> None:
        self.input_stream_handle = stream
        self.integration_stream = LiftedGroupAdd(self.input_stream_handle, None)
        self.delayed_stream = Delay(self.integration_stream.output_handle())
        self.integration_stream.set_input_b(self.delayed_stream.output_handle())

        self.output_stream_handle = self.integration_stream.output_handle()

    def step(self) -> bool:
        """
        Adds the latest element from the input stream to the running sum
        """
        self.delayed_stream.step()
        self.integration_stream.step()

        return self.output().current_time() == self.input_a().current_time()


class LiftedDelay(Lift1[Stream[T], Stream[T]]):
    """
    Lifts the Delay operator to work on streams of streams.
    """

    def __init__(self, stream: StreamHandle[Stream[T]]):
        super().__init__(
            stream,
            lambda s: step_until_fixpoint_and_return(Delay(StreamHandle(lambda: s))),
            None,
        )


class LiftedIntegrate(Lift1[Stream[T], Stream[T]]):
    """
    Lifts the Integrate operator to work on streams of streams.
    """

    def __init__(self, stream: StreamHandle[Stream[T]]):
        super().__init__(
            stream,
            lambda s: step_until_fixpoint_and_return(Integrate(StreamHandle(lambda: s))),
            None,
        )


class LiftedDifferentiate(Lift1[Stream[T], Stream[T]]):
    """
    Lifts the Differentiate operator to work on streams of streams.
    """

    def __init__(self, stream: StreamHandle[Stream[T]]):
        super().__init__(
            stream,
            lambda s: step_until_fixpoint_and_return(Differentiate(StreamHandle(lambda: s))),
            None,
        )


class LiftedStreamIntroduction(Lift1[T, Stream[T]]):
    """
    Lifts the stream_introduction function to work on streams.
    """

    def __init__(self, stream: StreamHandle[T]) -> None:
        group = stream.get().group()

        super().__init__(
            stream,
            lambda x: stream_introduction(x, group),
            StreamAddition(group),
        )


class LiftedStreamElimination(Lift1[Stream[T], T]):
    """
    Lifts the stream_elimination function to work on streams of streams.
    """

    def __init__(self, stream: StreamHandle[Stream[T]]) -> None:
        super().__init__(stream, lambda x: stream_elimination(x), stream.get().group().inner_group())  # type: ignore



from typing import Iterator, List

from pydbsp.core import AbelianGroupOperation
from pydbsp.zset import ZSet, ZSetAddition


class LazyZSet[T]:
    inner: List[ZSet[T]]

    def __init__(self, zsets: List[ZSet[T]]) -> None:
        self.inner = zsets

    def __iter__(self) -> Iterator[ZSet[T]]:
        """
        Returns an iterable of ZSets
        """
        return self.inner.__iter__()

    def __repr__(self) -> str:
        return self.inner.__repr__()

    def is_identity(self) -> bool:
        """
        Returns whether all ZSets are empty
        """
        return all(map(lambda zset: zset.is_identity(), self.inner))

    def coalesce(self) -> ZSet[T]:
        """
        Coalesces the LazyZSet into a single ZSet and returns it. This is a very expensive operation.

        If there is only a single ZSet however, it is cheap.
        """
        if len(self.inner) == 1:
            return self.inner[0]

        group = ZSetAddition[T]()
        coalesced_lazy_zset = group.identity()
        for zset in self:
            coalesced_lazy_zset = group.add(coalesced_lazy_zset, zset)

        return coalesced_lazy_zset

    def __eq__(self, other: object) -> bool:
        """
        Two LazyZSets are equal if their coalesced form match.

        This is a very expensive operation, as it requires coalescing all ZSets. It
        is however cheap to know whether a LazyZSet is the identity element, or if equality
        is being done against a identity element.
        """
        if not isinstance(other, LazyZSet):
            return False

        is_self_identity = self.is_identity()
        is_other_identity = other.is_identity()

        if is_self_identity and is_other_identity:
            return True
        elif is_self_identity and not is_other_identity:
            return False

        coalesced_self = self.coalesce()
        coalesced_other = other.coalesce()  # type: ignore

        return coalesced_self == coalesced_other

    def __contains__(self, item: T) -> bool:
        """
        An item is in the lazy Z-set if the sum of its weight across all ZSets contained within it is greater than zero.
        """
        return self[item] > 0

    def __getitem__(self, item: T) -> int:
        """
        Returns the sum of the weight of a given item across all ZSets.
        """
        weight = 0

        for zset in self.inner:
            weight += zset[item]

        return weight


class LazyZSetAddition[T](AbelianGroupOperation[LazyZSet[T]]):
    def add(self, a: LazyZSet[T], b: LazyZSet[T]) -> LazyZSet[T]:
        result = LazyZSet(a.inner + b.inner)
        return result

    def neg(self, a: LazyZSet[T]) -> LazyZSet[T]:
        group = ZSetAddition[T]()
        return LazyZSet([group.neg(zset) for zset in a])

    def identity(self) -> LazyZSet[T]:
        return LazyZSet([])
